import itertools
from typing import Any, List, Mapping, Tuple
from gymnasium import ActionWrapper, Env, ObservationWrapper, RewardWrapper, spaces, Wrapper
import numpy as np
import pandas as pd

try:
    from ray.rllib.env import MultiAgentEnv
except (ModuleNotFoundError, ImportError) as e:
    from gymnasium import Env as MultiAgentEnv

from citylearn.citylearn import CityLearnEnv
from citylearn.building import Building

class ClippedObservationWrapper(ObservationWrapper):
    """Wrapper for observations min-max and periodic normalization.
    
    Observations are clipped to be within the observation space limits.

    Parameters
    ----------
    env: CityLearnEnv
        CityLearn environment.
    """

    def __init__(self, env: CityLearnEnv) -> None:
        super().__init__(env)
        self.env: CityLearnEnv

    def observation(self, observations: List[List[float]]) -> List[List[float]]:
        """Returns normalized observations."""

        for i, (o, s) in enumerate(zip(observations, self.observation_space)):
            for j, (o_, l, u) in enumerate(zip(o, s.low, s.high)):
                observations[i][j] = min(max(o_, l), u)

        return observations

class NormalizedObservationWrapper(ObservationWrapper):
    """Wrapper for observations min-max and periodic normalization.
    
    Temporal observations including `hour`, `day_type` and `month` are periodically normalized using sine/cosine 
    transformations and then all observations are min-max normalized between 0 and 1.

    Parameters
    ----------
    env: CityLearnEnv
        CityLearn environment.
    """

    def __init__(self, env: CityLearnEnv) -> None:
        super().__init__(env)
        self.env: CityLearnEnv

    @property
    def shared_observations(self) -> List[str]:
        """Names of common observations across all buildings i.e. observations that have the same value irrespective of the building.
        
        Includes extra three observations added during cyclic transformation of :code:`hour`, :code:`day_type` and :code:`month`.
        """

        shared_observations = []
        periodic_observation_names = list(Building.get_periodic_observation_metadata().keys())

        for o in self.env.shared_observations:
            if o in periodic_observation_names:
                shared_observations += [f'{o}_cos', f'{o}_sin']
            
            else:
                shared_observations.append(o)

        return shared_observations
    
    @property
    def observation_names(self) -> List[List[str]]:
        """Names of returned observations.

        Includes extra three observations added during cyclic transformation of :code:`hour`, :code:`day_type` and :code:`month`.

        Notes
        -----
        If `central_agent` is True, a list of 1 sublist containing all building observation names is returned in the same order as `buildings`. 
        The `shared_observations` names are only included in the first building's observation names. If `central_agent` is False, a list of sublists 
        is returned where each sublist is a list of 1 building's observation names and the sublist in the same order as `buildings`.
        """

        if self.env.unwrapped.central_agent:
            observation_names = []

            for i, b in enumerate(self.env.unwrapped.buildings):
                for k, _ in b.observations(normalize=True, periodic_normalization=True).items():
                    if i == 0 or k not in self.shared_observations or k not in observation_names:
                        observation_names.append(k)
                    
                    else:
                        pass

            observation_names = [observation_names]
        
        else:
            observation_names = [list(b.observations(normalize=True, periodic_normalization=True).keys()) for b in self.env.buildings]

        return observation_names


    @property
    def observation_space(self) -> List[spaces.Box]:
        """Returns observation space for normalized observations."""

        low_limit = []
        high_limit = []

        if self.env.unwrapped.central_agent:
            shared_observations = []

            for i, b in enumerate(self.env.unwrapped.buildings):
                s = b.estimate_observation_space(normalize=True)
                o = b.observations(normalize=True, periodic_normalization=True)

                for k, lv, hv in zip(o, s.low, s.high):                    
                    if i == 0 or k not in self.shared_observations or k not in shared_observations:
                        low_limit.append(lv)
                        high_limit.append(hv)

                    else:
                        pass

                    if k in self.shared_observations and k not in shared_observations:
                        shared_observations.append(k)
                    
                    else:
                        pass
            
            observation_space = [spaces.Box(low=np.array(low_limit), high=np.array(high_limit), dtype=np.float32)]

        else:
            observation_space = [b.estimate_observation_space(normalize=True) for b in self.env.unwrapped.buildings]
        
        return observation_space

    def observation(self, observations: List[List[float]]) -> List[List[float]]:
        """Returns normalized observations."""

        if self.env.unwrapped.central_agent:
            norm_observations = []
            shared_observations = []

            for i, b in enumerate(self.env.unwrapped.buildings):
                for k, v in b.observations(normalize=True, periodic_normalization=True).items():
                    if i==0 or k not in self.shared_observations or k not in shared_observations:
                        norm_observations.append(v)

                    else:
                        pass

                    if k in self.shared_observations and k not in shared_observations:
                        shared_observations.append(k)
                    
                    else:
                        pass
            
            norm_observations = [norm_observations]

        else:
            norm_observations = [list(b.observations(normalize=True, periodic_normalization=True).values()) for b in self.env.unwrapped.buildings]
        
        return norm_observations
    
class NormalizedActionWrapper(ActionWrapper):
    """Wrapper for action min-max normalization.
    
    All observations are min-max normalized between 0 and 1.

    Parameters
    ----------
    env: CityLearnEnv
        CityLearn environment.
    """

    def __init__(self, env: CityLearnEnv) -> None:
        super().__init__(env)
        self.env: CityLearnEnv
        
    @property
    def action_space(self) -> List[spaces.Box]:
        """Returns action space for normalized actions."""

        low_limit = []
        high_limit = []

        if self.env.unwrapped.central_agent:

            for b in self.env.unwrapped.buildings:
                low_limit += [0.0]*b.action_space.low.size
                high_limit += [1.0]*b.action_space.high.size
            
            action_space = [spaces.Box(low=np.array(low_limit), high=np.array(high_limit), dtype=np.float32)]

        else:
            action_space = [spaces.Box(
                low=np.array([0.0]*b.action_space.low.size), 
                high=np.array([1.0]*b.action_space.high.size), 
                dtype=np.float32) 
            for b in self.env.unwrapped.buildings]
        
        return action_space

    def action(self, actions: List[float]) -> List[List[float]]:
        """Returns denormalized actions."""

        transformed_actions = []

        for i, s in enumerate(self.env.unwrapped.action_space):
            transformed_actions_ = []
            
            for j, (l, h) in enumerate(zip(s.low, s.high)):
                a = actions[i][j]*(h - l) + l
                transformed_actions_.append(a)
            
            transformed_actions.append(transformed_actions_)

        return transformed_actions
    
class NormalizedSpaceWrapper(Wrapper):
    """Wrapper for normalized observation and action spaces.

    Wraps `env` in :py:class:`citylearn.wrappers.NormalizedObservationWrapper` and :py:class:`citylearn.wrappers.NormalizedActionWrapper`.

    Parameters
    ----------
    env: CityLearnEnv
        CityLearn environment.
    """

    def __init__(self, env: CityLearnEnv):
        env = NormalizedObservationWrapper(env)
        env = NormalizedActionWrapper(env)
        super().__init__(env)
        self.env: CityLearnEnv
    
class DiscreteObservationWrapper(ObservationWrapper):
    """Wrapper for observation space discretization.

    Parameters
    ----------
    env: CityLearnEnv
        CityLearn environment.
    bin_sizes: List[Mapping[str, int]], optional
        Then number of bins for each active observation in each building.
    default_bin_size: int, default = 10
        The default number of bins if `bin_sizes` is unspecified for any active building observation.
    """

    def __init__(self, env: CityLearnEnv, bin_sizes: List[Mapping[str, int]] = None, default_bin_size: int = None):
        super().__init__(env)
        self.env: CityLearnEnv
        assert bin_sizes is None or len(bin_sizes) == len(self.env.unwrapped.buildings), 'length of bin_size must equal number of buildings.'
        self.bin_sizes = [{} for _ in self.env.unwrapped.buildings] if bin_sizes is None else bin_sizes
        self.default_bin_size = 10 if default_bin_size is None else default_bin_size
        self.bin_sizes = [
            {o: s.get(o, self.default_bin_size) for o in b.active_observations} 
            for b, s in zip(self.env.unwrapped.buildings, self.bin_sizes)
        ]
        
    @property
    def observation_space(self) -> List[spaces.MultiDiscrete]:
        """Returns observation space for discretized observations."""

        if self.env.unwrapped.central_agent:
            bin_sizes = []
            shared_observations = []

            for i, b in enumerate(self.bin_sizes):
                for k, v in b.items():
                    if i == 0 or k not in self.env.shared_observations or k not in shared_observations:
                        bin_sizes.append(v)

                    else:
                        pass

                    if k in self.env.shared_observations and k not in shared_observations:
                        shared_observations.append(k)

                    else:
                        pass
            
            observation_space = [spaces.MultiDiscrete(bin_sizes)]

        else:
            observation_space = [spaces.MultiDiscrete(list(b.values())) for b in self.bin_sizes]
        
        return observation_space
    
    def observation(self, observations: List[List[float]]) -> np.ndarray:
        """Returns discretized observations."""

        transformed_observations = []

        for i, (cs, ds) in enumerate(zip(self.env.unwrapped.observation_space, self.observation_space)):
            transformed_observations_ = []

            for j, (ll, hl, b) in enumerate(zip(cs.low, cs.high, ds)):
                o = np.digitize(observations[i][j], np.linspace(ll, hl, b.n), right=True)
                transformed_observations_.append(o)

            transformed_observations.append(transformed_observations_)
                
        return transformed_observations
    
class DiscreteActionWrapper(ActionWrapper):
    """Wrapper for action space discretization.

    Parameters
    ----------
    env: CityLearnEnv
        CityLearn environment.
    bin_sizes: List[Mapping[str, int]], optional
        Then number of bins for each active action in each building.
    default_bin_size: int, default = 10
        The default number of bins if `bin_sizes` is unspecified for any active building action.
    """

    def __init__(self, env: CityLearnEnv, bin_sizes: List[Mapping[str, int]] = None, default_bin_size: int = None):
        super().__init__(env)
        self.env: CityLearnEnv
        assert bin_sizes is None or len(bin_sizes) == len(self.env.unwrapped.buildings), 'length of bin_size must equal number of buildings.'
        self.bin_sizes = [{} for _ in self.env.unwrapped.buildings] if bin_sizes is None else bin_sizes
        self.default_bin_size = 10 if default_bin_size is None else default_bin_size
        self.bin_sizes = [
            {a: s.get(a, self.default_bin_size) for a in b.active_actions} 
            for b, s in zip(self.env.unwrapped.buildings, self.bin_sizes)
        ]
        
    @property
    def action_space(self) -> List[spaces.MultiDiscrete]:
        """Returns action space for discretized actions."""

        if self.env.unwrapped.central_agent:
            bin_sizes = []

            for b in self.bin_sizes:
                for _, v in b.items():
                    bin_sizes.append(v)
            
            action_space = [spaces.MultiDiscrete(bin_sizes)]

        else:
            action_space = [spaces.MultiDiscrete(list(b.values())) for b in self.bin_sizes]

        return action_space

    def action(self, actions: List[float]) -> List[List[float]]:
        """Returns undiscretized actions."""

        transformed_actions = []

        for i, (cs, ds) in enumerate(zip(self.env.unwrapped.action_space, self.action_space)):
            transformed_actions_ = []
            
            for j, (ll, hl, b) in enumerate(zip(cs.low, cs.high, ds)):
                a = np.linspace(ll, hl, b.n)[actions[i][j]]
                transformed_actions_.append(a)
            
            transformed_actions.append(transformed_actions_)

        return transformed_actions
    
class DiscreteSpaceWrapper(Wrapper):
    """Wrapper for observation and action spaces discretization.

    Wraps `env` in :py:class:`citylearn.wrappers.DiscreteObservationWrapper` and :py:class:`citylearn.wrappers.DiscreteActionWrapper`.

    Parameters
    ----------
    env: CityLearnEnv
        CityLearn environment.
    observation_bin_sizes: List[Mapping[str, int]], optional
        Then number of bins for each active observation in each building.
    action_bin_sizes: List[Mapping[str, int]], optional
        Then number of bins for each active action in each building.
    default_observation_bin_size: int, default = 10
        The default number of bins if `bin_sizes` is unspecified for any active building observation.
    default_action_bin_size: int, default = 10
        The default number of bins if `bin_sizes` is unspecified for any active building action.
    """

    def __init__(self, env: CityLearnEnv, observation_bin_sizes: List[Mapping[str, int]] = None, action_bin_sizes: List[Mapping[str, int]] = None, default_observation_bin_size: int = None, default_action_bin_size: int = None):
        env = DiscreteObservationWrapper(env, bin_sizes=observation_bin_sizes, default_bin_size=default_observation_bin_size)
        env = DiscreteActionWrapper(env, bin_sizes=action_bin_sizes, default_bin_size=default_action_bin_size)
        super().__init__(env)
        self.env: CityLearnEnv

class TabularQLearningObservationWrapper(ObservationWrapper):
    """Observation wrapper for :py:class:`citylearn.agents.q_learning.TabularQLearning` agent.

    Wraps `env` in :py:class:`citylearn.wrappers.DiscreteObservationWrapper`.
    
    Parameters
    ----------
    env: CityLearnEnv
        CityLearn environment.
    bin_sizes: List[Mapping[str, int]], optional
        Then number of bins for each active observation in each building.
    default_bin_size: int, default = 10
        The default number of bins if `bin_sizes` is unspecified for any active building observation.
    """

    def __init__(self, env: CityLearnEnv, bin_sizes: List[Mapping[str, int]] = None, default_bin_size: int = None) -> None:
        env = DiscreteObservationWrapper(env, bin_sizes=bin_sizes, default_bin_size=default_bin_size)
        super().__init__(env)
        self.env: CityLearnEnv
        self.combinations = self.set_combinations()

    @property
    def observation_space(self) -> List[spaces.Discrete]:
        """Returns observation space for discretized observations."""

        observation_space = []

        for c in self.combinations:
            observation_space.append(spaces.Discrete(len(c) - 1))
        
        return observation_space
    
    def observation(self, observations: List[List[int]]) -> List[List[int]]:
        """Returns discretized observations."""

        return [[c.index(tuple(o))] for o, c in zip(observations, self.combinations)]
    
    def set_combinations(self) -> List[List[int]]:
        """Returns all combinations of discrete observations."""

        combs_list = []

        for s in self.env.observation_space:
            options = [list(range(d.n + 1)) for d in s]
            combs = list(itertools.product(*options))
            combs_list.append(combs)

        return combs_list
    
class TabularQLearningActionWrapper(ActionWrapper):
    """Action wrapper for :py:class:`citylearn.agents.q_learning.TabularQLearning` agent.

    Wraps `env` in :py:class:`citylearn.wrappers.DiscreteActionWrapper`.
    
    Parameters
    ----------
    env: CityLearnEnv
        CityLearn environment.
    bin_sizes: List[Mapping[str, int]], optional
        Then number of bins for each active action in each building.
    default_bin_size: int, default = 10
        The default number of bins if `bin_sizes` is unspecified for any active building action.
    """

    def __init__(self, env: CityLearnEnv, bin_sizes: List[Mapping[str, int]] = None, default_bin_size: int = None) -> None:
        env = DiscreteActionWrapper(env, bin_sizes=bin_sizes, default_bin_size=default_bin_size)
        super().__init__(env)
        self.env: CityLearnEnv
        self.combinations = self.set_combinations()

    @property
    def action_space(self) -> List[spaces.Discrete]:
        """Returns action space for discretized actions."""

        action_space = []

        for c in self.combinations:
            action_space.append(spaces.Discrete(len(c)))
        
        return action_space
    
    def action(self, actions: List[float]) -> List[List[int]]:
        """Returns discretized actions."""

        return [list(c[a[0]]) for a, c in zip(actions, self.combinations)]
    
    def set_combinations(self) -> List[List[int]]:
        """Returns all combinations of discrete actions."""

        combs_list = []

        for s in self.env.action_space:
            options = [list(range(d.n)) for d in s]
            combs = list(itertools.product(*options))
            combs_list.append(combs)

        return combs_list
    
class TabularQLearningWrapper(Wrapper):
    """Wrapper for :py:class:`citylearn.agents.q_learning.TabularQLearning` agent.

    Discretizes observation and action spaces.

    Parameters
    ----------
    env: CityLearnEnv
        CityLearn environment.
    observation_bin_sizes: List[Mapping[str, int]], optional
        Then number of bins for each active observation in each building.
    action_bin_sizes: List[Mapping[str, int]], optional
        Then number of bins for each active action in each building.
    default_observation_bin_size: int, default = 10
        The default number of bins if `bin_sizes` is unspecified for any active building observation.
    default_action_bin_size: int, default = 10
        The default number of bins if `bin_sizes` is unspecified for any active building action.
    """

    def __init__(self, env: CityLearnEnv, observation_bin_sizes: List[Mapping[str, int]] = None, action_bin_sizes: List[Mapping[str, int]] = None, default_observation_bin_size: int = None, default_action_bin_size: int = None):
        env = TabularQLearningObservationWrapper(env, bin_sizes=observation_bin_sizes, default_bin_size=default_observation_bin_size)
        env = TabularQLearningActionWrapper(env, bin_sizes=action_bin_sizes, default_bin_size=default_action_bin_size)
        super().__init__(env)
        self.env: CityLearnEnv

class StableBaselines3ObservationWrapper(ObservationWrapper):
    """Observation wrapper for :code:`stable-baselines3` algorithms.

    Wraps observations so that they are returned in a 1-dimensional numpy array.
    This wrapper is only compatible when the environment is controlled by a central agent
    i.e., :py:attr:`citylearn.citylearn.CityLearnEnv.central_agent` = True.
    
    Parameters
    ----------
    env: CityLearnEnv
        CityLearn environment.
    """

    def __init__(self, env: CityLearnEnv):
        assert env.unwrapped.central_agent, 'StableBaselines3ObservationWrapper is compatible only when env.central_agent = True.'\
            ' First set env.central_agent = True to use this wrapper.'
        
        super().__init__(env)
        self.env: CityLearnEnv
        
    @property
    def observation_space(self) -> spaces.Box:
        """Returns single spaces Box object."""

        return self.env.observation_space[0]
    
    def observation(self, observations: List[List[float]]) -> np.ndarray:
        """Returns observations as 1-dimensional numpy array."""

        return np.array(observations[0], dtype='float32')

class StableBaselines3ActionWrapper(ActionWrapper):
    """Action wrapper for :code:`stable-baselines3` algorithms.

    Wraps actions so that they are returned in a 1-dimensional numpy array.
    This wrapper is only compatible when the environment is controlled by a central agent
    i.e., :py:attr:`citylearn.citylearn.CityLearnEnv.central_agent` = True.
    
    Parameters
    ----------
    env: CityLearnEnv
        CityLearn environment.
    """

    def __init__(self, env: CityLearnEnv):
        assert env.unwrapped.central_agent, 'StableBaselines3ActionWrapper is compatible only when env.central_agent = True.'\
            ' First set env.central_agent = True to use this wrapper.'
        
        super().__init__(env)
        self.env: CityLearnEnv

    @property
    def action_space(self) -> spaces.Box:
        """Returns single spaces Box object."""

        return self.env.action_space[0]

    def action(self, actions: List[float]) -> List[List[float]]:
        """Returns actions as 1-dimensional numpy array."""

        return [actions]
    
class StableBaselines3RewardWrapper(RewardWrapper):
    """Reward wrapper for :code:`stable-baselines3` algorithms.

    Wraps rewards so that it is returned as float value.
    This wrapper is only compatible when the environment is controlled by a central agent
    i.e., :py:attr:`citylearn.citylearn.CityLearnEnv.central_agent` = True.
    
    Parameters
    ----------
    env: CityLearnEnv
        CityLearn environment.
    """

    def __init__(self, env: CityLearnEnv):
        assert env.unwrapped.central_agent, 'StableBaselines3RewardWrapper is compatible only when env.central_agent = True.'\
            ' First set env.central_agent = True to use this wrapper.'
        
        super().__init__(env)
        self.env: CityLearnEnv

    def reward(self, reward: List[float]) -> float:
        """Returns reward as float value."""

        return reward[0]

class StableBaselines3Wrapper(Wrapper):
    """Wrapper for :code:`stable-baselines3` algorithms.

    Wraps observations so that they are returned in a 1-dimensional numpy array.
    Wraps actions so that they are returned in a 1-dimensional numpy array.
    Wraps rewards so that it is returned as float value.
    
    Parameters
    ----------
    env: CityLearnEnv
        CityLearn environment.
    """

    def __init__(self, env: CityLearnEnv):
        env = StableBaselines3ActionWrapper(env)
        env = StableBaselines3RewardWrapper(env)
        env = StableBaselines3ObservationWrapper(env)
        super().__init__(env)
        self.env: CityLearnEnv

class RLlibSingleAgentWrapper(StableBaselines3Wrapper):
    """Wrapper for :code:`RLlib` single-agent algorithms.

    Uses the same wrapper as :code:`stable-baselines3` by wrapping
    `env` in :py:class:`citylearn.wrappers.StableBaselines3Wrapper`.
    
    Parameters
    ----------
    env_config: Mapping[str, Any]
        Dictionary providing initialization parameters for the environment.
        Must contain `env_kwargs` as a key where `env_kwargs` is a `dict` used to 
        initialize :py:class:`citylearn.citylearn.CityLearnEnv`. Thus it must 
        contain all with positional arguments needed for intialization and 
        optionally contain optional  intialization arguments. `env_config` can also 
        contain a `wrappers` key that is a list of :py:mod:`citylearn.wrappers` 
        classes to wrap :py:class:`citylearn.citylearn.CityLearnEnv` with. Wrapping
        with :py:class:`citylearn.wrappers.ClippedObservationWrapper` is recommended
        to avoid having the simulation terminating prematurely with an error due to
        out of bound observations relative to the observation space.

    Notes
    -----
    This wrapper is only compatible with an environment where 
    :py:attr:`citylearn.citylearn.central_agent`=`True` and will initialize the 
    environment as such, overriding any value for `central_agent` in `env_kwargs`.
    """

    def __init__(self, env_config: Mapping[str, Any]):
        env_kwargs = env_config['env_kwargs']
        env_kwargs['central_agent'] = True
        assert 'schema' in env_kwargs, 'missing schema key in env_kwargs.'
        env = CityLearnEnv(**env_kwargs)

        wrappers = env_config.get('wrappers')
        wrappers = [] if wrappers is None else wrappers

        for w in wrappers:
            env = w(env)

        super().__init__(env)

class RLlibMultiAgentObservationWrapper(ObservationWrapper):
    """Observation wrapper for :code:`RLlib` multi-agent algorithms.

    Wraps observation space and observations so that they are returned 
    as :py:class:`gymnasium.spaces.Dict` and `dict` objects respectively.
    The keys in these objects correspond to the agent IDs i.e., 
    policy IDs in the multi-agent.
    
    Parameters
    ----------
    env: CityLearnEnv
        CityLearn environment.
    """

    def __init__(self, env: CityLearnEnv):
        assert not env.central_agent, 'RLlibMultiAgentObservationWrapper is'\
            ' compatible only when env.central_agent = False.'\
                ' First set env.central_agent = False to use this wrapper.'

        super().__init__(env)
        self.env: CityLearnEnv

    @property
    def observation_space(self) -> spaces.Dict:
        """Parses observation space into a :py:class:`gymnasium.spaces.Dict`."""

        return spaces.Dict({f'agent_{i}': s for i, s in enumerate(self.env.observation_space)})

    def observation(
        self, observations: List[List[float]]
    ) -> Mapping[str, np.ndarray]:
        """Parses observation into a dictionary."""

        return {f'agent_{i}': np.array(o, dtype='float32') for i, o in enumerate(observations)}

class RLlibMultiAgentActionWrapper(ActionWrapper):
    """Action wrapper for :code:`RLlib` multi-agent algorithms.

    Wraps action space so that it is returned as :py:class:`gymnasium.spaces.Dict`. 
    The keys correspond to the agent IDs i.e., policy IDs in the multi-agent. 
    Also converts agent actions from `dict` to  data structure need by 
    :py:meth:`citylearn.citylearn.CityLearnEnv.step`.
    
    Parameters
    ----------
    env: CityLearnEnv
        CityLearn environment.
    """

    def __init__(self, env: CityLearnEnv):
        assert not env.central_agent, 'RLlibMultiAgentActionWrapper is'\
            ' compatible only when env.central_agent = False.'\
                ' First set env.central_agent = False to use this wrapper.'

        super().__init__(env)
        self.env: CityLearnEnv

    @property
    def action_space(self) -> spaces.Dict:
        """Parses action space into a :py:class:`gymnasium.spaces.Dict`."""

        return spaces.Dict({f'agent_{i}': s for i, s in enumerate(self.env.action_space)})

    def action(self, actions: Mapping[str, np.ndarray]) -> List[List[float]]:
        """Parses actions into data structure for :py:meth:`citylearn.citylearn.CityLearnEnv.step`."""

        return [list(v) for v in actions.values()]

class RLlibMultiAgentRewardWrapper(RewardWrapper):
    """Action wrapper for :code:`RLlib` multi-agent algorithms.

    Wraps action space so that it is returned as a `dict` mapping agent IDs to reward values.
    
    Parameters
    ----------
    env: CityLearnEnv
        CityLearn environment.
    """

    def __init__(self, env: CityLearnEnv):
        assert not env.central_agent, 'RLlibMultiAgentRewardWrapper is'\
            ' compatible only when env.central_agent = False.'\
                ' First set env.central_agent = False to use this wrapper.'

        super().__init__(env)
        self.env: CityLearnEnv

    def reward(self, reward: List[float]) -> Mapping[str, float]:
        """Parses reward into a `dict`."""

        return {f'agent_{i}': r for i, r in enumerate(reward)}

class RLlibMultiAgentEnv(MultiAgentEnv):
    """Wrapper for :code:`RLlib` multi-agent algorithms.

    Converts, observation and action spaces to :py:class:`gymnasium.spaces.Dict`.
    Also converts `observations`, `actions`, `rewards`, `terminated`, and `truncated`
    to dictionaries where necessary. The dictionary keys correspond to the agent IDs i.e., 
    policy IDs in the multi-agent. Agent IDs are accessible through the `_agent_ids` property. 
    The initialized environment is a :py:class:`ray.rllib.env.MultiAgentEnv` object 
    and has an `env` attribute that is :py:class:`citylearn.citylearn.CityLearnEnv` object.
    
    Parameters
    ----------
    env_config: Mapping[str, Any]
        Dictionary providing initialization parameters for the environment.
        Must contain `env_kwargs` as a key where `env_kwargs` is a `dict` used to 
        initialize :py:class:`citylearn.citylearn.CityLearnEnv`. Thus it must 
        contain all with positional arguments needed for intialization and 
        optionally contain optional  intialization arguments. `env_config` can also 
        contain a `wrappers` key that is a list of :py:mod:`citylearn.wrappers` 
        classes to wrap :py:class:`citylearn.citylearn.CityLearnEnv` with. Wrapping
        with :py:class:`citylearn.wrappers.ClippedObservationWrapper` is recommended
        to avoid having the simulation terminating prematurely with an error due to
        out of bound observations relative to the observation space.

    Notes
    -----
    This wrapper is only compatible with an environment where 
    :py:attr:`citylearn.citylearn.central_agent`=`False` and will initialize the 
    environment as such, overriding any value for `central_agent` in `env_kwargs`.
    """
    
    def __init__(self, env_config: Mapping[str, Any]):
        if MultiAgentEnv == Env:
            raise Exception('This functionality requires you to install RLlib.'\
                'You can install RLlib from pip: pip install "ray[rllib]", '\
                    'or for more detailed instructions please visit https://docs.ray.io/en/latest/rllib/index.html.')
        
        else:
            pass

        super().__init__()
        env_kwargs = env_config['env_kwargs']
        env_kwargs['central_agent'] = False
        assert 'schema' in env_kwargs, 'missing schema key in env_kwargs.'
        env = CityLearnEnv(**env_kwargs)

        wrappers = env_config.get('wrappers')
        wrappers = [] if wrappers is None else wrappers

        for w in wrappers:
            env = w(env)

        env = RLlibMultiAgentActionWrapper(env)
        env = RLlibMultiAgentObservationWrapper(env)
        env = RLlibMultiAgentRewardWrapper(env)
        self.env: CityLearnEnv = env
        self._agent_ids = [f'agent_{i}' for i in range(len(self.buildings))]
        self.observation_space: spaces.Dict = self.env.observation_space
        self.action_space: spaces.Dict = self.env.action_space

    @property
    def time_step(self) -> int:
        """Convenience property for :py:meth:`citylearn.citylearn.CityLearnEnv.time_step`."""

        return self.env.unwrapped.time_step

    @property
    def buildings(self) -> List[Building]:
        """Convenience property for :py:meth:`citylearn.citylearn.CityLearnEnv.buildings`."""

        return self.env.unwrapped.buildings
    
    @property
    def terminated(self) -> bool:
        """Convenience property for :py:meth:`citylearn.citylearn.CityLearnEnv.terminated`."""

        return self.env.unwrapped.terminated

    def step(
            self, action_dict: Mapping[str, np.ndarray]
    ) -> Tuple[Mapping[str, np.ndarray], Mapping[str, float], Mapping[str, bool], Mapping[str, bool], Mapping[str, dict]]:
        """Calls :py:meth:`citylearn.citylearn.CityLearnEnv.step` and parses returned values into dictionaries."""

        observations, reward, terminated, truncated, info = self.env.step(action_dict)
        terminated = {'__all__': terminated, **{a: terminated for a in self._agent_ids}}
        truncated = {'__all__': truncated, **{a: truncated for a in self._agent_ids}}
        info = {a: info for a in self._agent_ids}

        return observations, reward, terminated, truncated, info

    def evaluate(self, **kwargs) -> pd.DataFrame:
        """Convenience method for :py:meth:`citylearn.citylearn.CityLearnEnv.evaluate`."""

        return self.env.unwrapped.evaluate(**kwargs)

    def reset(self, *, seed: int = None, options: Mapping[str, Any] = None) -> Tuple[Mapping[str, np.ndarray], Mapping[str, dict]]:
        """Calls :py:meth:`citylearn.citylearn.CityLearnEnv.reset` and parses returned values into dictionaries."""

        observations, info = self.env.reset(seed=seed, options=options)
        info = {a: info for a in self._agent_ids}

        return observations, info


from typing import Any, List
import numpy as np
from citylearn.citylearn import CityLearnEnv
from citylearn.preprocessing import Encoder, PeriodicNormalization, Normalize, OnehotEncoding

# conditional imports
try:
    import torch
except (ModuleNotFoundError, ImportError) as e:
    raise Exception("This functionality requires you to install torch. You can install torch by : pip install torch torchvision, or for more detailed instructions please visit https://pytorch.org.")

from citylearn.agents.base import Agent

class RLC(Agent):
    r"""Base reinforcement learning controller class.

    Parameters
    ----------
    env: CityLearnEnv
        CityLearn environment.
    hidden_dimension : List[float], default: [256, 256]
        Hidden dimension.
    discount : float, default: 0.99
        Discount factor.
    tau : float, default: 5e-3
        Decay rate.
    alpha: float, default: 0.2
        Temperature; exploration-exploitation balance term.
    lr : float, default: 3e-4
        Learning rate.
    batch_size : int, default: 256
        Batch size.
    replay_buffer_capacity : int, default: 1e5
        Replay buffer capacity.
    standardize_start_time_step : int, optional
        Time step to calculate mean and standard deviation, and begin standardization of observations and rewards in replay buffer. Defaults to :py:attr:`citylearn.citylearn.CityLearnEnv.time_steps` - 2.
    end_exploration_time_step : int, optional
        Time step to stop random or RBC-guided exploration. Defaults to :py:attr:`citylearn.citylearn.CityLearnEnv.time_steps` - 1.
    action_scaling_coefficient : float, default: 0.5
        Action scaling coefficient.
    reward_scaling : float, default: 5.0
        Reward scaling.
    update_per_time_step : int, default: 2
        Number of updates per time step.
    
    Other Parameters
    ----------------
    **kwargs : Any
        Other keyword arguments used to initialize super class.
    """
            
    def __init__(
        self, env: CityLearnEnv, hidden_dimension: List[float] = None, 
        discount: float = None, tau: float = None, alpha: float = None, lr: float = None, batch_size: int = None,
        replay_buffer_capacity: int = None, standardize_start_time_step: int = None, end_exploration_time_step: int = None, 
        action_scaling_coefficienct: float = None, reward_scaling: float = None, 
        update_per_time_step: int = None, **kwargs: Any
    ):
        super().__init__(env, **kwargs)
        self.hidden_dimension = hidden_dimension
        self.discount = discount
        self.tau = tau
        self.alpha = alpha
        self.lr = lr
        self.batch_size = batch_size
        self.replay_buffer_capacity = replay_buffer_capacity
        self.standardize_start_time_step = standardize_start_time_step
        self.end_exploration_time_step = end_exploration_time_step
        self.action_scaling_coefficient = action_scaling_coefficienct
        self.reward_scaling = reward_scaling
        self.update_per_time_step = update_per_time_step
        self.encoders = self.set_encoders()

    @property
    def observation_dimension(self) -> int:
        """Number of observations after applying `encoders`."""

        return [len([j for j in np.hstack(e*np.ones(len(s.low))) if j != None]) for e, s in zip(self.encoders, self.observation_space)]

    @property
    def hidden_dimension(self) -> List[float]:
        """Hidden dimension."""

        return self.__hidden_dimension

    @property
    def discount(self) -> float:
        """Discount factor."""

        return self.__discount

    @property
    def tau(self) -> float:
        """Decay rate."""

        return self.__tau

    @property
    def alpha(self) -> float:
        """Temperature; exploration-exploitation balance term."""

        return self.__alpha
    
    @property
    def lr(self) -> float:
        """Learning rate."""

        return self.__lr

    @property
    def batch_size(self) -> int:
        """Batch size."""

        return self.__batch_size

    @property
    def replay_buffer_capacity(self) -> int:
        """Replay buffer capacity."""

        return self.__replay_buffer_capacity

    @property
    def standardize_start_time_step(self) -> int:
        """Time step to calculate mean and standard deviation, and begin standardization of observations and rewards in replay buffer. Defaults to :py:attr:`citylearn.citylearn.CityLearnEnv.time_steps` - 2."""

        return self.__standardize_start_time_step

    @property
    def end_exploration_time_step(self) -> int:
        """Time step to stop exploration.  Defaults to :py:attr:`citylearn.citylearn.CityLearnEnv.time_steps` - 1."""

        return self.__end_exploration_time_step

    @property
    def action_scaling_coefficient(self) -> float:
        """Action scaling coefficient."""

        return self.__action_scaling_coefficient

    @property
    def reward_scaling(self) -> float:
        """Reward scaling."""

        return self.__reward_scaling

    @property
    def update_per_time_step(self) -> int:
        """Number of updates per time step."""

        return self.__update_per_time_step

    @hidden_dimension.setter
    def hidden_dimension(self,  hidden_dimension: List[float]):
        self.__hidden_dimension = [256, 256] if hidden_dimension is None else hidden_dimension

    @discount.setter
    def discount(self, discount: float):
        self.__discount = 0.99 if discount is None else discount

    @tau.setter
    def tau(self, tau: float):
        self.__tau = 5e-3 if tau is None else tau
    
    @alpha.setter
    def alpha(self, alpha: float):
        self.__alpha = 0.2 if alpha is None else alpha
    
    @lr.setter
    def lr(self, lr: float):
        self.__lr = 3e-4 if lr is None else lr

    @batch_size.setter
    def batch_size(self, batch_size: int):
        self.__batch_size = 256 if batch_size is None else batch_size

    @replay_buffer_capacity.setter
    def replay_buffer_capacity(self, replay_buffer_capacity: int):
        self.__replay_buffer_capacity = 1e5 if replay_buffer_capacity is None else replay_buffer_capacity

    @standardize_start_time_step.setter
    def standardize_start_time_step(self, standardize_start_time_step: int):
        self.__standardize_start_time_step = self.episode_time_steps - 1 if standardize_start_time_step is None else standardize_start_time_step

    @end_exploration_time_step.setter
    def end_exploration_time_step(self, end_exploration_time_step: int):
        self.__end_exploration_time_step = self.episode_time_steps if end_exploration_time_step is None else end_exploration_time_step

    @action_scaling_coefficient.setter
    def action_scaling_coefficient(self, action_scaling_coefficient: float):
        self.__action_scaling_coefficient = 0.5 if action_scaling_coefficient is None else action_scaling_coefficient

    @reward_scaling.setter
    def reward_scaling(self, reward_scaling: float):
        self.__reward_scaling = 5.0 if reward_scaling is None else reward_scaling

    @update_per_time_step.setter
    def update_per_time_step(self, update_per_time_step: int):
        update_per_time_step = 2 if update_per_time_step is None else update_per_time_step
        assert isinstance(update_per_time_step,int), f'update_per_time_step mut be int type. {update_per_time_step} is of {type(update_per_time_step)} type'
        self.__update_per_time_step = update_per_time_step

    @Agent.random_seed.setter
    def random_seed(self, seed: int):
        Agent.random_seed.fset(self, seed)
        torch.manual_seed(self.random_seed)

    def set_encoders(self) -> List[List[Encoder]]:
        r"""Get observation value transformers/encoders for use in agent algorithm.

        The encoder classes are defined in the `preprocessing.py` module and include `PeriodicNormalization` for cyclic observations,
        `OnehotEncoding` for categorical obeservations, `RemoveFeature` for non-applicable observations given available storage systems and devices
        and `Normalize` for observations with known minimum and maximum boundaries.
        
        Returns
        -------
        encoders : List[List[Encoder]]
            Encoder classes for observations ordered with respect to `active_observations`.
        """

        encoders = []

        for o, s in zip(self.observation_names, self.observation_space):
            e = []

            for i, n in enumerate(o):
                if n in ['month', 'hour']:
                    e.append(PeriodicNormalization(s.high[i]))
            
                elif n == 'day_type':
                    e.append(OnehotEncoding([1, 2, 3, 4, 5, 6, 7, 8]))
            
                elif n == "daylight_savings_status":
                    e.append(OnehotEncoding([0, 1]))
            
                else:
                    e.append(Normalize(s.low[i], s.high[i]))

            encoders.append(e)

        return encoders


from pathlib import Path

import numpy as np
import pytest

pytest.importorskip("gymnasium")

from citylearn.citylearn import CityLearnEnv, EvaluationCondition
from citylearn.cost_function import CostFunction
from citylearn.data import ZERO_DIVISION_PLACEHOLDER


SCHEMA = Path(__file__).resolve().parents[1] / "data/datasets/citylearn_challenge_2022_phase_all_plus_evs/schema.json"


def _run_episode(schema: Path, action_value: float, episode_time_steps: int = 48) -> CityLearnEnv:
    env = CityLearnEnv(
        str(schema),
        central_agent=True,
        episode_time_steps=episode_time_steps,
        random_seed=0,
    )
    env.reset()
    ev_indices = [
        idx
        for idx, name in enumerate(env.action_names[0])
        if name.startswith("electric_vehicle_storage_")
    ]

    base_action = np.zeros(env.action_space[0].shape[0], dtype="float32")

    while not env.terminated:
        action = base_action.copy()
        if ev_indices:
            action[ev_indices] = action_value

        env.step([action])

    return env


def test_kpi_normalization_matches_baseline():
    env = _run_episode(SCHEMA, action_value=0.0)

    try:
        df = env.evaluate(
            control_condition=EvaluationCondition.WITH_STORAGE_AND_PARTIAL_LOAD_AND_PV,
            baseline_condition=EvaluationCondition.WITH_STORAGE_AND_PARTIAL_LOAD_AND_PV,
        )

        normalized_keys = {
            "ramping_average",
            "daily_one_minus_load_factor_average",
            "monthly_one_minus_load_factor_average",
            "daily_peak_average",
            "all_time_peak_average",
        }
        district = df[(df["level"] == "district") & (df["cost_function"].isin(normalized_keys))]
        vals = district["value"].to_numpy(dtype=float)

        assert np.all(np.isfinite(vals)), "District KPI contains non-finite values."
        assert np.allclose(vals, 1.0, atol=1e-5), "District KPI normalization deviates when control==baseline."
    finally:
        env.close()


def test_ev_charging_load_impacts_building_kpis():
    def _total_charger_consumption(sim_env: CityLearnEnv) -> float:
        return float(sum(np.sum(b.chargers_electricity_consumption) for b in sim_env.buildings))

    base_env = _run_episode(SCHEMA, action_value=0.0, episode_time_steps=24)
    try:
        base_df = base_env.evaluate()
        base_charger_consumption = _total_charger_consumption(base_env)
    finally:
        base_env.close()

    charged_env = _run_episode(SCHEMA, action_value=1.0, episode_time_steps=24)
    try:
        charged_df = charged_env.evaluate()
        charged_charger_consumption = _total_charger_consumption(charged_env)
    finally:
        charged_env.close()

    def _extract(df):
        subset = df[
            (df["level"] == "building")
            & (df["cost_function"] == "electricity_consumption_total")
        ]
        return subset.set_index("name")["value"]

    base_values = _extract(base_df)
    charged_values = _extract(charged_df)

    charged_values = charged_values.reindex(base_values.index)

    assert charged_charger_consumption > base_charger_consumption, \
        "EV load should increase charger electricity consumption."
    assert np.isfinite(charged_values.dropna()).all()


@pytest.mark.parametrize("seconds_per_time_step", [5, 10, 60, 300, 900])
def test_histories_and_kpi_consistency_with_subhour_steps(seconds_per_time_step: int):
    env = CityLearnEnv(
        str(SCHEMA),
        central_agent=True,
        episode_time_steps=6,
        seconds_per_time_step=seconds_per_time_step,
        random_seed=0,
    )

    try:
        env.reset()
        names = env.action_names[0]
        base_action = np.zeros(env.action_space[0].shape[0], dtype="float32")

        ev_indices = [idx for idx, name in enumerate(names) if name.startswith("electric_vehicle_storage_")]
        battery_indices = [idx for idx, name in enumerate(names) if "electrical_storage" in name]

        while not env.terminated:
            action = base_action.copy()
            if ev_indices:
                action[ev_indices] = 0.8
            if battery_indices:
                action[battery_indices] = 0.5

            env.step([action])
            t = env.time_step - 1

            assert 1 <= len(env.net_electricity_consumption) <= env.time_step + 1
            assert 1 <= len(env.net_electricity_consumption_cost) <= env.time_step + 1
            assert 1 <= len(env.net_electricity_consumption_emission) <= env.time_step + 1

            for building in env.buildings:
                lhs = building.net_electricity_consumption[t]
                rhs = (
                    building.cooling_electricity_consumption[t]
                    + building.heating_electricity_consumption[t]
                    + building.dhw_electricity_consumption[t]
                    + building.non_shiftable_load_electricity_consumption[t]
                    + building.electrical_storage_electricity_consumption[t]
                    + building.solar_generation[t]
                    + building.chargers_electricity_consumption[t]
                    + building.washing_machines_electricity_consumption[t]
                )
                assert abs(lhs - rhs) < 1e-4

        final_t = env.time_step
        committed_len = final_t if final_t > 0 else 1
        for building in env.buildings:
            if building.electric_vehicle_chargers:
                charger_total = float(np.sum(building.chargers_electricity_consumption))
                charger_components = float(
                    sum(np.sum(c.electricity_consumption[: final_t + 1]) for c in building.electric_vehicle_chargers)
                )
                assert charger_total == pytest.approx(charger_components)

            assert len(building.solar_generation) == committed_len
            assert np.all(np.isfinite(building.solar_generation))

        for ev in env.electric_vehicles:
            soc = ev.battery.soc[: committed_len]
            assert np.all(np.isfinite(soc))
            assert np.all((soc >= 0.0) & (soc <= 1.0))

        control = EvaluationCondition.WITH_STORAGE_AND_PV
        baseline = EvaluationCondition.WITHOUT_STORAGE_BUT_WITH_PV

        def _safe_div(control_value: float, baseline_value: float):
            eps = float(ZERO_DIVISION_PLACEHOLDER)
            if abs(baseline_value) <= eps:
                return 1.0 if abs(control_value) <= eps else np.nan
            return control_value / baseline_value

        building_ratios = []
        for building in env.buildings:
            ec_c = CostFunction.electricity_consumption(
                np.array(getattr(building, f"net_electricity_consumption{control.value}"), dtype=float).tolist()
            )[-1]
            ec_b = CostFunction.electricity_consumption(
                np.array(getattr(building, f"net_electricity_consumption{baseline.value}"), dtype=float).tolist()
            )[-1]
            building_ratios.append(_safe_div(float(ec_c), float(ec_b)))

        expected_ratio = float(np.nanmean(np.array(building_ratios, dtype=float)))

        df = env.evaluate(control_condition=control, baseline_condition=baseline)
        district_value = float(
            df[
                (df["level"] == "district")
                & (df["cost_function"] == "electricity_consumption_total")
            ]["value"].iloc[0]
        )

        assert district_value == pytest.approx(expected_ratio)
    finally:
        env.close()



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Mapping, Optional, Tuple

import numpy as np
import pandas as pd

from citylearn.cost_function import CostFunction
from citylearn.data import EnergySimulation, ZERO_DIVISION_PLACEHOLDER

if TYPE_CHECKING:
    from citylearn.citylearn import CityLearnEnv


class CityLearnKPIService:
    """Internal KPI/evaluation service for `CityLearnEnv`."""

    def __init__(self, env: "CityLearnEnv"):
        self.env = env

    @staticmethod
    def _to_scalar(value, default: float = 0.0) -> float:
        try:
            scalar = float(value)
        except (TypeError, ValueError):
            return float(default)

        if not np.isfinite(scalar):
            return float(default)

        return scalar

    @staticmethod
    def _safe_div(control_value: float, baseline_value: float):
        c = CityLearnKPIService._to_scalar(control_value, 0.0)
        b = CityLearnKPIService._to_scalar(baseline_value, 0.0)
        eps = float(ZERO_DIVISION_PLACEHOLDER)

        if abs(b) <= eps:
            return 1.0 if abs(c) <= eps else None

        return c / b

    @staticmethod
    def _window_steps(window_seconds: float, seconds_per_time_step: float) -> int:
        step_seconds = max(float(seconds_per_time_step), 1.0)
        return max(1, int(round(float(window_seconds) / step_seconds)))

    @staticmethod
    def _simulated_days(env: "CityLearnEnv") -> float:
        steps = max(int(getattr(env, 'time_step', 0)), 1)
        step_seconds = max(float(getattr(env, 'seconds_per_time_step', 0) or 0), 1.0)
        return (steps * step_seconds) / (24.0 * 3600.0)

    @staticmethod
    def _daily_average(total_value: float, simulated_days: float) -> Optional[float]:
        value = CityLearnKPIService._to_scalar(total_value, np.nan)

        if not np.isfinite(value):
            return None

        if simulated_days <= float(ZERO_DIVISION_PLACEHOLDER):
            return None

        return float(value / simulated_days)

    @staticmethod
    def _normalize_soc_target(value) -> Optional[float]:
        try:
            target = float(value)
        except (TypeError, ValueError):
            return None

        if not np.isfinite(target):
            return None

        if target > 1.0 and target <= 100.0:
            target = target / 100.0

        if target < 0.0 or target > 1.0:
            return None

        return float(target)

    @staticmethod
    def _metric(cost_function: str, value, name: str, level: str) -> Dict[str, object]:
        return {
            'cost_function': cost_function,
            'value': value,
            'name': name,
            'level': level,
        }

    @staticmethod
    def _sum_finite(values) -> float:
        try:
            series = np.array(values, dtype='float64').flatten()
        except (TypeError, ValueError):
            return 0.0

        if series.size == 0:
            return 0.0

        finite = series[np.isfinite(series)]
        if finite.size == 0:
            return 0.0

        return float(finite.sum())

    @staticmethod
    def _equity_relative_benefit_percent(cost_scenario: float, cost_baseline: float) -> Optional[float]:
        scenario = CityLearnKPIService._to_scalar(cost_scenario, np.nan)
        baseline = CityLearnKPIService._to_scalar(cost_baseline, np.nan)

        if not np.isfinite(scenario) or not np.isfinite(baseline) or baseline <= 0.0:
            return None

        return float(100.0 * (baseline - scenario) / baseline)

    @staticmethod
    def _equity_distribution_metrics(relative_benefits: np.ndarray) -> Dict[str, Optional[float]]:
        benefits = np.array(relative_benefits, dtype='float64')
        benefits = benefits[np.isfinite(benefits)]

        if benefits.size == 0:
            return {
                'equity_gini_benefit': None,
                'equity_cr20_benefit': None,
                'equity_losers_percent': None,
            }

        losers_percent = float(100.0 * np.count_nonzero(benefits < 0.0) / benefits.size)
        benefits_plus = np.clip(benefits, 0.0, None)
        total_plus = float(benefits_plus.sum())

        if total_plus <= 0.0:
            return {
                'equity_gini_benefit': None,
                'equity_cr20_benefit': None,
                'equity_losers_percent': losers_percent,
            }

        n = benefits_plus.size
        diff_sum = float(np.abs(benefits_plus[:, None] - benefits_plus[None, :]).sum())
        gini = float(diff_sum / (2.0 * n * total_plus))

        k = max(1, int(np.ceil(0.2 * n)))
        top_sum = float(np.sort(benefits_plus)[::-1][:k].sum())
        cr20 = float(top_sum / total_plus)

        return {
            'equity_gini_benefit': gini,
            'equity_cr20_benefit': cr20,
            'equity_losers_percent': losers_percent,
        }

    @staticmethod
    def _equity_bpr(
        non_negative_relative_benefits: Mapping[str, float],
        groups: Mapping[str, Optional[str]],
    ) -> Optional[float]:
        if len(non_negative_relative_benefits) == 0:
            return None

        asset_poor_values = []
        asset_rich_values = []

        for building_name, value in non_negative_relative_benefits.items():
            group = groups.get(building_name)

            if group == 'asset_poor':
                asset_poor_values.append(float(value))
            elif group == 'asset_rich':
                asset_rich_values.append(float(value))
            else:
                return None

        if len(asset_poor_values) == 0 or len(asset_rich_values) == 0:
            return None

        rich_mean = float(np.mean(asset_rich_values))
        poor_mean = float(np.mean(asset_poor_values))

        if rich_mean <= 0.0:
            return None

        return float(poor_mean / rich_mean)

    def _compute_ev_metrics(self, building) -> Dict[str, float]:
        t_final = int(max(building.time_step, 0))
        departures_total = 0
        departures_met = 0
        departure_deficit_sum = 0.0
        charge_total_kwh = 0.0
        v2g_export_total_kwh = 0.0

        for charger in building.electric_vehicle_chargers or []:
            consumption = np.array(charger.electricity_consumption[0:t_final + 1], dtype='float64')
            charge_total_kwh += float(np.clip(consumption, 0.0, None).sum())
            v2g_export_total_kwh += float(np.clip(-consumption, 0.0, None).sum())

            sim = charger.charger_simulation
            states = np.array(sim.electric_vehicle_charger_state, dtype='float64')
            required_soc = np.array(sim.electric_vehicle_required_soc_departure, dtype='float64')
            history_limit = min(t_final, len(states) - 2, len(required_soc) - 1, len(charger.past_connected_evs) - 1)

            if history_limit < 0:
                continue

            for t in range(history_limit + 1):
                current_state = states[t]
                next_state = states[t + 1]

                if current_state != 1 or next_state == 1:
                    continue

                ev = charger.past_connected_evs[t]
                if ev is None:
                    continue

                target_soc = self._normalize_soc_target(required_soc[t])
                if target_soc is None:
                    continue

                if t >= len(ev.battery.soc):
                    continue

                actual_soc = self._to_scalar(ev.battery.soc[t], np.nan)
                if not np.isfinite(actual_soc):
                    continue

                departures_total += 1
                deficit = max(target_soc - actual_soc, 0.0)
                departure_deficit_sum += deficit
                if deficit <= 1e-6:
                    departures_met += 1

        success_rate = None if departures_total == 0 else departures_met / departures_total
        deficit_mean = None if departures_total == 0 else departure_deficit_sum / departures_total

        return {
            'departures_total': float(departures_total),
            'departures_met': float(departures_met),
            'departure_deficit_sum': float(departure_deficit_sum),
            'ev_departure_success_rate': success_rate,
            'ev_departure_soc_deficit_mean': deficit_mean,
            'ev_charge_total_kwh': float(charge_total_kwh),
            'ev_v2g_export_total_kwh': float(v2g_export_total_kwh),
        }

    def _compute_bess_metrics(self, building) -> Dict[str, float]:
        t_final = int(max(building.time_step, 0))
        storage = building.electrical_storage
        storage_series = np.array(building.electrical_storage_electricity_consumption[0:t_final + 1], dtype='float64')
        charge_total = float(np.clip(storage_series, 0.0, None).sum())
        discharge_total = float(np.clip(-storage_series, 0.0, None).sum())
        throughput_total = charge_total + discharge_total

        capacity = self._to_scalar(getattr(storage, 'capacity', 0.0), 0.0)
        degraded_capacity = self._to_scalar(getattr(storage, 'degraded_capacity', capacity), capacity)
        equivalent_cycles = None if capacity <= 0.0 else throughput_total / (2.0 * capacity)
        fade_ratio = None if capacity <= 0.0 else (capacity - degraded_capacity) / capacity

        if fade_ratio is not None:
            fade_ratio = float(np.clip(fade_ratio, 0.0, 1.0))

        return {
            'bess_charge_total_kwh': charge_total,
            'bess_discharge_total_kwh': discharge_total,
            'bess_throughput_total_kwh': throughput_total,
            'bess_equivalent_full_cycles': equivalent_cycles,
            'bess_capacity_fade_ratio': fade_ratio,
            '_bess_capacity_kwh': capacity,
            '_bess_degraded_capacity_kwh': degraded_capacity,
        }

    def _compute_pv_metrics(self, building) -> Dict[str, float]:
        t_final = int(max(building.time_step, 0))
        solar = np.array(building.solar_generation[0:t_final + 1], dtype='float64')
        net = np.array(building.net_electricity_consumption[0:t_final + 1], dtype='float64')

        generation = np.clip(-solar, 0.0, None)
        export = np.clip(-net, 0.0, None)
        pv_generation_total = float(generation.sum())
        pv_export_total = float(np.minimum(generation, export).sum())
        self_consumption_ratio = None if pv_generation_total <= 0.0 else (pv_generation_total - pv_export_total) / pv_generation_total

        return {
            'pv_generation_total_kwh': pv_generation_total,
            'pv_export_total_kwh': pv_export_total,
            'pv_self_consumption_ratio': self_consumption_ratio,
        }

    def _compute_phase_metrics(self, building) -> Dict[str, object]:
        if not getattr(building, '_electrical_service_enabled', False):
            return {
                'electrical_service_violation_total_kwh': 0.0,
                'electrical_service_violation_time_step_count': 0.0,
                'phase_imbalance_ratio_average': None,
                'phase_import_peak_kw': {},
                'phase_export_peak_kw': {},
                '_imbalance_sum': 0.0,
                '_imbalance_count': 0.0,
            }

        t_final = int(max(building.time_step, 0))
        violation_history = np.array(getattr(building, '_charging_constraint_violation_history', [0.0]), dtype='float64')[0:t_final + 1]
        violation_total = float(np.clip(violation_history, 0.0, None).sum())
        violation_count = float(np.count_nonzero(violation_history > 1e-9))

        phase_history = getattr(building, '_charging_phase_power_history_kw', {}) or {}
        phase_import_peak = {}
        phase_export_peak = {}

        for phase_name, values in phase_history.items():
            series = np.array(values[0:t_final + 1], dtype='float64')
            phase_import_peak[phase_name] = float(np.clip(series, 0.0, None).max(initial=0.0))
            phase_export_peak[phase_name] = float(np.clip(-series, 0.0, None).max(initial=0.0))

        imbalance_sum = 0.0
        imbalance_count = 0.0
        imbalance_average = None

        if getattr(building, '_electrical_service_mode', 'single_phase') == 'three_phase':
            names = [n for n in ['L1', 'L2', 'L3'] if n in phase_history]
            if len(names) == 3:
                stacked = np.stack([np.array(phase_history[n][0:t_final + 1], dtype='float64') for n in names], axis=1)
                for row in stacked:
                    abs_row = np.abs(row)
                    mean_abs = float(abs_row.mean())
                    if mean_abs <= 1e-9:
                        ratio = 0.0
                    else:
                        ratio = float((abs_row.max() - abs_row.min()) / mean_abs)
                    imbalance_sum += ratio
                    imbalance_count += 1.0

                if imbalance_count > 0:
                    imbalance_average = imbalance_sum / imbalance_count

        return {
            'electrical_service_violation_total_kwh': violation_total,
            'electrical_service_violation_time_step_count': violation_count,
            'phase_imbalance_ratio_average': imbalance_average,
            'phase_import_peak_kw': phase_import_peak,
            'phase_export_peak_kw': phase_export_peak,
            '_imbalance_sum': imbalance_sum,
            '_imbalance_count': imbalance_count,
        }

    def _collect_market_totals(self, building_names: List[str]) -> Tuple[Mapping[str, Mapping[str, float]], Mapping[str, float]]:
        history = getattr(self.env, '_community_market_settlement_history', []) or []
        by_building = {
            name: {
                'community_local_import_total_kwh': 0.0,
                'community_local_export_total_kwh': 0.0,
                'community_grid_import_after_local_total_kwh': 0.0,
                'community_grid_export_after_local_total_kwh': 0.0,
                'community_settled_cost_total_eur': 0.0,
                'community_counterfactual_cost_total_eur': 0.0,
                'community_market_savings_total_eur': 0.0,
            }
            for name in building_names
        }

        for rows in history:
            for row in rows:
                name = row.get('building')
                if name not in by_building:
                    continue

                target = by_building[name]
                target['community_local_import_total_kwh'] += self._to_scalar(row.get('local_import_kwh'), 0.0)
                target['community_local_export_total_kwh'] += self._to_scalar(row.get('local_export_kwh'), 0.0)
                target['community_grid_import_after_local_total_kwh'] += self._to_scalar(row.get('grid_import_kwh'), 0.0)
                target['community_grid_export_after_local_total_kwh'] += self._to_scalar(row.get('grid_export_kwh'), 0.0)
                target['community_settled_cost_total_eur'] += self._to_scalar(row.get('settled_cost_eur', row.get('settled_cost')), 0.0)
                target['community_counterfactual_cost_total_eur'] += self._to_scalar(row.get('counterfactual_cost_eur'), 0.0)
                target['community_market_savings_total_eur'] += self._to_scalar(row.get('market_savings_eur'), 0.0)

        district = {
            key: float(sum(values[key] for values in by_building.values()))
            for key in [
                'community_local_import_total_kwh',
                'community_local_export_total_kwh',
                'community_grid_import_after_local_total_kwh',
                'community_grid_export_after_local_total_kwh',
                'community_settled_cost_total_eur',
                'community_counterfactual_cost_total_eur',
                'community_market_savings_total_eur',
            ]
        }

        return by_building, district

    def evaluate(
        self,
        control_condition=None,
        baseline_condition=None,
        comfort_band: float = None,
        *,
        evaluation_condition_cls,
        dynamics_building_cls,
    ) -> pd.DataFrame:
        """Evaluate cost functions at current time step."""

        env = self.env

        get_net_electricity_consumption = lambda x, c: getattr(x, f'net_electricity_consumption{c.value}')
        get_net_electricity_consumption_cost = lambda x, c: getattr(x, f'net_electricity_consumption_cost{c.value}')
        get_net_electricity_consumption_emission = lambda x, c: getattr(x, f'net_electricity_consumption_emission{c.value}')

        comfort_band = EnergySimulation.DEFUALT_COMFORT_BAND if comfort_band is None else comfort_band
        daily_steps = self._window_steps(24.0 * 3600.0, env.seconds_per_time_step)
        monthly_steps = self._window_steps(730.0 * 3600.0, env.seconds_per_time_step)
        simulated_days = self._simulated_days(env)

        legacy_building_frames: List[pd.DataFrame] = []
        extended_building_rows: List[Dict[str, object]] = []

        ev_departures_total = 0.0
        ev_departures_met = 0.0
        ev_deficit_sum = 0.0
        ev_charge_total = 0.0
        ev_v2g_total = 0.0

        bess_charge_total = 0.0
        bess_discharge_total = 0.0
        bess_throughput_total = 0.0
        bess_capacity_total = 0.0
        bess_capacity_loss_total = 0.0

        pv_generation_total = 0.0
        pv_export_total = 0.0

        phase_violation_total = 0.0
        phase_violation_count = 0.0
        phase_imbalance_sum = 0.0
        phase_imbalance_count = 0.0

        building_names = [building.name for building in env.buildings]
        equity_group_by_building = {building.name: getattr(building, 'equity_group', None) for building in env.buildings}
        equity_relative_benefit_by_building: Dict[str, Optional[float]] = {}
        equity_valid_benefits: Dict[str, float] = {}

        for building in env.buildings:
            if isinstance(building, dynamics_building_cls):
                building_control_condition = (
                    evaluation_condition_cls.WITH_STORAGE_AND_PARTIAL_LOAD_AND_PV
                    if control_condition is None else control_condition
                )
                building_baseline_condition = (
                    evaluation_condition_cls.WITHOUT_STORAGE_AND_PARTIAL_LOAD_BUT_WITH_PV
                    if baseline_condition is None else baseline_condition
                )
            else:
                building_control_condition = (
                    evaluation_condition_cls.WITH_STORAGE_AND_PV
                    if control_condition is None else control_condition
                )
                building_baseline_condition = (
                    evaluation_condition_cls.WITHOUT_STORAGE_BUT_WITH_PV
                    if baseline_condition is None else baseline_condition
                )

            discomfort_kwargs = {
                'indoor_dry_bulb_temperature': building.indoor_dry_bulb_temperature,
                'dry_bulb_temperature_cooling_set_point': building.indoor_dry_bulb_temperature_cooling_set_point,
                'dry_bulb_temperature_heating_set_point': building.indoor_dry_bulb_temperature_heating_set_point,
                'band': building.comfort_band if comfort_band is None else comfort_band,
                'occupant_count': building.occupant_count,
            }
            unmet, cold, hot, \
                cold_minimum_delta, cold_maximum_delta, cold_average_delta, \
                hot_minimum_delta, hot_maximum_delta, hot_average_delta = CostFunction.discomfort(**discomfort_kwargs)
            expected_energy = building.cooling_demand + building.heating_demand + building.dhw_demand + building.non_shiftable_load
            served_energy = building.energy_from_cooling_device + building.energy_from_cooling_storage \
                + building.energy_from_heating_device + building.energy_from_heating_storage \
                + building.energy_from_dhw_device + building.energy_from_dhw_storage \
                + building.energy_to_non_shiftable_load
            ec_c = CostFunction.electricity_consumption(get_net_electricity_consumption(building, building_control_condition))[-1]
            ec_b = CostFunction.electricity_consumption(get_net_electricity_consumption(building, building_baseline_condition))[-1]
            zne_c = CostFunction.zero_net_energy(get_net_electricity_consumption(building, building_control_condition))[-1]
            zne_b = CostFunction.zero_net_energy(get_net_electricity_consumption(building, building_baseline_condition))[-1]
            ce_c = CostFunction.carbon_emissions(get_net_electricity_consumption_emission(building, building_control_condition))[-1]
            ce_b = CostFunction.carbon_emissions(get_net_electricity_consumption_emission(building, building_baseline_condition))[-1] if sum(building.carbon_intensity.carbon_intensity) != 0 else 0
            control_cost_series = get_net_electricity_consumption_cost(building, building_control_condition)
            baseline_cost_series = get_net_electricity_consumption_cost(building, building_baseline_condition)
            cost_c_legacy = CostFunction.cost(control_cost_series)[-1]
            cost_b_legacy = CostFunction.cost(baseline_cost_series)[-1]
            cost_c_raw = self._sum_finite(control_cost_series)
            cost_b_raw = self._sum_finite(baseline_cost_series)
            equity_benefit = self._equity_relative_benefit_percent(cost_c_raw, cost_b_raw)
            equity_relative_benefit_by_building[building.name] = equity_benefit

            if equity_benefit is not None:
                equity_valid_benefits[building.name] = float(equity_benefit)

            legacy_building_frame = pd.DataFrame([{
                'cost_function': 'electricity_consumption_total',
                'value': self._safe_div(ec_c, ec_b),
            }, {
                'cost_function': 'zero_net_energy',
                'value': self._safe_div(zne_c, zne_b),
            }, {
                'cost_function': 'carbon_emissions_total',
                'value': self._safe_div(ce_c, ce_b),
            }, {
                'cost_function': 'cost_total',
                'value': self._safe_div(cost_c_legacy, cost_b_legacy),
            }, {
                'cost_function': 'discomfort_proportion',
                'value': unmet[-1],
            }, {
                'cost_function': 'discomfort_cold_proportion',
                'value': cold[-1],
            }, {
                'cost_function': 'discomfort_hot_proportion',
                'value': hot[-1],
            }, {
                'cost_function': 'discomfort_cold_delta_minimum',
                'value': cold_minimum_delta[-1],
            }, {
                'cost_function': 'discomfort_cold_delta_maximum',
                'value': cold_maximum_delta[-1],
            }, {
                'cost_function': 'discomfort_cold_delta_average',
                'value': cold_average_delta[-1],
            }, {
                'cost_function': 'discomfort_hot_delta_minimum',
                'value': hot_minimum_delta[-1],
            }, {
                'cost_function': 'discomfort_hot_delta_maximum',
                'value': hot_maximum_delta[-1],
            }, {
                'cost_function': 'discomfort_hot_delta_average',
                'value': hot_average_delta[-1],
            }, {
                'cost_function': 'one_minus_thermal_resilience_proportion',
                'value': CostFunction.one_minus_thermal_resilience(power_outage=building.power_outage_signal, **discomfort_kwargs)[-1],
            }, {
                'cost_function': 'power_outage_normalized_unserved_energy_total',
                'value': CostFunction.normalized_unserved_energy(expected_energy, served_energy, power_outage=building.power_outage_signal)[-1],
            }, {
                'cost_function': 'annual_normalized_unserved_energy_total',
                'value': CostFunction.normalized_unserved_energy(expected_energy, served_energy)[-1],
            }])
            legacy_building_frame['name'] = building.name
            legacy_building_frames.append(legacy_building_frame)

            extended_building_rows.extend([
                self._metric('electricity_consumption_control_total_kwh', ec_c, building.name, 'building'),
                self._metric('electricity_consumption_baseline_total_kwh', ec_b, building.name, 'building'),
                self._metric('electricity_consumption_delta_total_kwh', ec_c - ec_b, building.name, 'building'),
                self._metric('electricity_consumption_control_daily_average_kwh', self._daily_average(ec_c, simulated_days), building.name, 'building'),
                self._metric('electricity_consumption_baseline_daily_average_kwh', self._daily_average(ec_b, simulated_days), building.name, 'building'),
                self._metric('electricity_consumption_delta_daily_average_kwh', self._daily_average(ec_c - ec_b, simulated_days), building.name, 'building'),
                self._metric('zero_net_energy_control_total_kwh', zne_c, building.name, 'building'),
                self._metric('zero_net_energy_baseline_total_kwh', zne_b, building.name, 'building'),
                self._metric('zero_net_energy_delta_total_kwh', zne_c - zne_b, building.name, 'building'),
                self._metric('zero_net_energy_control_daily_average_kwh', self._daily_average(zne_c, simulated_days), building.name, 'building'),
                self._metric('zero_net_energy_baseline_daily_average_kwh', self._daily_average(zne_b, simulated_days), building.name, 'building'),
                self._metric('zero_net_energy_delta_daily_average_kwh', self._daily_average(zne_c - zne_b, simulated_days), building.name, 'building'),
                self._metric('carbon_emissions_control_total_kgco2', ce_c, building.name, 'building'),
                self._metric('carbon_emissions_baseline_total_kgco2', ce_b, building.name, 'building'),
                self._metric('carbon_emissions_delta_total_kgco2', ce_c - ce_b, building.name, 'building'),
                self._metric('carbon_emissions_control_daily_average_kgco2', self._daily_average(ce_c, simulated_days), building.name, 'building'),
                self._metric('carbon_emissions_baseline_daily_average_kgco2', self._daily_average(ce_b, simulated_days), building.name, 'building'),
                self._metric('carbon_emissions_delta_daily_average_kgco2', self._daily_average(ce_c - ce_b, simulated_days), building.name, 'building'),
                self._metric('cost_control_total_eur', cost_c_raw, building.name, 'building'),
                self._metric('cost_baseline_total_eur', cost_b_raw, building.name, 'building'),
                self._metric('cost_delta_total_eur', cost_c_raw - cost_b_raw, building.name, 'building'),
                self._metric('cost_control_daily_average_eur', self._daily_average(cost_c_raw, simulated_days), building.name, 'building'),
                self._metric('cost_baseline_daily_average_eur', self._daily_average(cost_b_raw, simulated_days), building.name, 'building'),
                self._metric('cost_delta_daily_average_eur', self._daily_average(cost_c_raw - cost_b_raw, simulated_days), building.name, 'building'),
                self._metric('equity_relative_benefit_percent', equity_benefit, building.name, 'building'),
            ])

            ev_metrics = self._compute_ev_metrics(building)
            extended_building_rows.extend([
                self._metric('ev_departure_success_rate', ev_metrics['ev_departure_success_rate'], building.name, 'building'),
                self._metric('ev_departure_soc_deficit_mean', ev_metrics['ev_departure_soc_deficit_mean'], building.name, 'building'),
                self._metric('ev_charge_total_kwh', ev_metrics['ev_charge_total_kwh'], building.name, 'building'),
                self._metric('ev_v2g_export_total_kwh', ev_metrics['ev_v2g_export_total_kwh'], building.name, 'building'),
            ])
            ev_departures_total += ev_metrics['departures_total']
            ev_departures_met += ev_metrics['departures_met']
            ev_deficit_sum += ev_metrics['departure_deficit_sum']
            ev_charge_total += ev_metrics['ev_charge_total_kwh']
            ev_v2g_total += ev_metrics['ev_v2g_export_total_kwh']

            bess_metrics = self._compute_bess_metrics(building)
            extended_building_rows.extend([
                self._metric('bess_charge_total_kwh', bess_metrics['bess_charge_total_kwh'], building.name, 'building'),
                self._metric('bess_discharge_total_kwh', bess_metrics['bess_discharge_total_kwh'], building.name, 'building'),
                self._metric('bess_throughput_total_kwh', bess_metrics['bess_throughput_total_kwh'], building.name, 'building'),
                self._metric('bess_equivalent_full_cycles', bess_metrics['bess_equivalent_full_cycles'], building.name, 'building'),
                self._metric('bess_capacity_fade_ratio', bess_metrics['bess_capacity_fade_ratio'], building.name, 'building'),
            ])
            bess_charge_total += bess_metrics['bess_charge_total_kwh']
            bess_discharge_total += bess_metrics['bess_discharge_total_kwh']
            bess_throughput_total += bess_metrics['bess_throughput_total_kwh']
            bess_capacity_total += bess_metrics['_bess_capacity_kwh']
            bess_capacity_loss_total += max(bess_metrics['_bess_capacity_kwh'] - bess_metrics['_bess_degraded_capacity_kwh'], 0.0)

            pv_metrics = self._compute_pv_metrics(building)
            extended_building_rows.extend([
                self._metric('pv_generation_total_kwh', pv_metrics['pv_generation_total_kwh'], building.name, 'building'),
                self._metric('pv_export_total_kwh', pv_metrics['pv_export_total_kwh'], building.name, 'building'),
                self._metric('pv_generation_daily_average_kwh', self._daily_average(pv_metrics['pv_generation_total_kwh'], simulated_days), building.name, 'building'),
                self._metric('pv_export_daily_average_kwh', self._daily_average(pv_metrics['pv_export_total_kwh'], simulated_days), building.name, 'building'),
                self._metric('pv_self_consumption_ratio', pv_metrics['pv_self_consumption_ratio'], building.name, 'building'),
            ])
            pv_generation_total += pv_metrics['pv_generation_total_kwh']
            pv_export_total += pv_metrics['pv_export_total_kwh']

            phase_metrics = self._compute_phase_metrics(building)
            extended_building_rows.extend([
                self._metric('electrical_service_violation_total_kwh', phase_metrics['electrical_service_violation_total_kwh'], building.name, 'building'),
                self._metric('electrical_service_violation_time_step_count', phase_metrics['electrical_service_violation_time_step_count'], building.name, 'building'),
                self._metric('phase_imbalance_ratio_average', phase_metrics['phase_imbalance_ratio_average'], building.name, 'building'),
            ])
            for phase_name, value in phase_metrics['phase_import_peak_kw'].items():
                extended_building_rows.append(self._metric(f'phase_import_peak_kw_{phase_name}', value, building.name, 'building'))
            for phase_name, value in phase_metrics['phase_export_peak_kw'].items():
                extended_building_rows.append(self._metric(f'phase_export_peak_kw_{phase_name}', value, building.name, 'building'))

            phase_violation_total += phase_metrics['electrical_service_violation_total_kwh']
            phase_violation_count += phase_metrics['electrical_service_violation_time_step_count']
            phase_imbalance_sum += phase_metrics['_imbalance_sum']
            phase_imbalance_count += phase_metrics['_imbalance_count']

        legacy_building = pd.concat(legacy_building_frames, ignore_index=True) if legacy_building_frames else pd.DataFrame(columns=['cost_function', 'value', 'name'])
        legacy_building['level'] = 'building'

        env_control_condition = (
            evaluation_condition_cls.WITH_STORAGE_AND_PARTIAL_LOAD_AND_PV
            if control_condition is None else control_condition
        )
        env_baseline_condition = (
            evaluation_condition_cls.WITHOUT_STORAGE_AND_PARTIAL_LOAD_BUT_WITH_PV
            if baseline_condition is None else baseline_condition
        )

        ramp_c = CostFunction.ramping(get_net_electricity_consumption(env, env_control_condition))[-1]
        ramp_b = CostFunction.ramping(get_net_electricity_consumption(env, env_baseline_condition))[-1]
        dlf_daily_c = CostFunction.one_minus_load_factor(get_net_electricity_consumption(env, env_control_condition), window=daily_steps)[-1]
        dlf_daily_b = CostFunction.one_minus_load_factor(get_net_electricity_consumption(env, env_baseline_condition), window=daily_steps)[-1]
        dlf_monthly_c = CostFunction.one_minus_load_factor(get_net_electricity_consumption(env, env_control_condition), window=monthly_steps)[-1]
        dlf_monthly_b = CostFunction.one_minus_load_factor(get_net_electricity_consumption(env, env_baseline_condition), window=monthly_steps)[-1]
        peak_daily_c = CostFunction.peak(get_net_electricity_consumption(env, env_control_condition), window=daily_steps)[-1]
        peak_daily_b = CostFunction.peak(get_net_electricity_consumption(env, env_baseline_condition), window=daily_steps)[-1]
        peak_all_c = CostFunction.peak(get_net_electricity_consumption(env, env_control_condition), window=env.time_steps)[-1]
        peak_all_b = CostFunction.peak(get_net_electricity_consumption(env, env_baseline_condition), window=env.time_steps)[-1]

        legacy_district_base = pd.DataFrame([{
            'cost_function': 'ramping_average',
            'value': self._safe_div(ramp_c, ramp_b),
        }, {
            'cost_function': 'daily_one_minus_load_factor_average',
            'value': self._safe_div(dlf_daily_c, dlf_daily_b),
        }, {
            'cost_function': 'monthly_one_minus_load_factor_average',
            'value': self._safe_div(dlf_monthly_c, dlf_monthly_b),
        }, {
            'cost_function': 'daily_peak_average',
            'value': self._safe_div(peak_daily_c, peak_daily_b),
        }, {
            'cost_function': 'all_time_peak_average',
            'value': self._safe_div(peak_all_c, peak_all_b),
        }])

        legacy_district = pd.concat([legacy_district_base, legacy_building], ignore_index=True, sort=False)
        legacy_district = legacy_district.groupby(['cost_function'])[['value']].mean().reset_index()
        legacy_district['name'] = 'District'
        legacy_district['level'] = 'district'

        legacy_cost_functions = pd.concat([legacy_district, legacy_building], ignore_index=True, sort=False)

        # Extended KPI district-level
        ec_c_env = CostFunction.electricity_consumption(get_net_electricity_consumption(env, env_control_condition))[-1]
        ec_b_env = CostFunction.electricity_consumption(get_net_electricity_consumption(env, env_baseline_condition))[-1]
        zne_c_env = CostFunction.zero_net_energy(get_net_electricity_consumption(env, env_control_condition))[-1]
        zne_b_env = CostFunction.zero_net_energy(get_net_electricity_consumption(env, env_baseline_condition))[-1]
        ce_c_env = CostFunction.carbon_emissions(get_net_electricity_consumption_emission(env, env_control_condition))[-1]
        ce_b_env = CostFunction.carbon_emissions(get_net_electricity_consumption_emission(env, env_baseline_condition))[-1]
        env_control_cost_series = get_net_electricity_consumption_cost(env, env_control_condition)
        env_baseline_cost_series = get_net_electricity_consumption_cost(env, env_baseline_condition)
        cost_c_env_raw = self._sum_finite(env_control_cost_series)
        cost_b_env_raw = self._sum_finite(env_baseline_cost_series)

        extended_district_rows = [
            self._metric('electricity_consumption_control_total_kwh', ec_c_env, 'District', 'district'),
            self._metric('electricity_consumption_baseline_total_kwh', ec_b_env, 'District', 'district'),
            self._metric('electricity_consumption_delta_total_kwh', ec_c_env - ec_b_env, 'District', 'district'),
            self._metric('electricity_consumption_control_daily_average_kwh', self._daily_average(ec_c_env, simulated_days), 'District', 'district'),
            self._metric('electricity_consumption_baseline_daily_average_kwh', self._daily_average(ec_b_env, simulated_days), 'District', 'district'),
            self._metric('electricity_consumption_delta_daily_average_kwh', self._daily_average(ec_c_env - ec_b_env, simulated_days), 'District', 'district'),
            self._metric('zero_net_energy_control_total_kwh', zne_c_env, 'District', 'district'),
            self._metric('zero_net_energy_baseline_total_kwh', zne_b_env, 'District', 'district'),
            self._metric('zero_net_energy_delta_total_kwh', zne_c_env - zne_b_env, 'District', 'district'),
            self._metric('zero_net_energy_control_daily_average_kwh', self._daily_average(zne_c_env, simulated_days), 'District', 'district'),
            self._metric('zero_net_energy_baseline_daily_average_kwh', self._daily_average(zne_b_env, simulated_days), 'District', 'district'),
            self._metric('zero_net_energy_delta_daily_average_kwh', self._daily_average(zne_c_env - zne_b_env, simulated_days), 'District', 'district'),
            self._metric('carbon_emissions_control_total_kgco2', ce_c_env, 'District', 'district'),
            self._metric('carbon_emissions_baseline_total_kgco2', ce_b_env, 'District', 'district'),
            self._metric('carbon_emissions_delta_total_kgco2', ce_c_env - ce_b_env, 'District', 'district'),
            self._metric('carbon_emissions_control_daily_average_kgco2', self._daily_average(ce_c_env, simulated_days), 'District', 'district'),
            self._metric('carbon_emissions_baseline_daily_average_kgco2', self._daily_average(ce_b_env, simulated_days), 'District', 'district'),
            self._metric('carbon_emissions_delta_daily_average_kgco2', self._daily_average(ce_c_env - ce_b_env, simulated_days), 'District', 'district'),
            self._metric('cost_control_total_eur', cost_c_env_raw, 'District', 'district'),
            self._metric('cost_baseline_total_eur', cost_b_env_raw, 'District', 'district'),
            self._metric('cost_delta_total_eur', cost_c_env_raw - cost_b_env_raw, 'District', 'district'),
            self._metric('cost_control_daily_average_eur', self._daily_average(cost_c_env_raw, simulated_days), 'District', 'district'),
            self._metric('cost_baseline_daily_average_eur', self._daily_average(cost_b_env_raw, simulated_days), 'District', 'district'),
            self._metric('cost_delta_daily_average_eur', self._daily_average(cost_c_env_raw - cost_b_env_raw, simulated_days), 'District', 'district'),
        ]

        ev_success_rate = None if ev_departures_total <= 0.0 else ev_departures_met / ev_departures_total
        ev_deficit_mean = None if ev_departures_total <= 0.0 else ev_deficit_sum / ev_departures_total
        extended_district_rows.extend([
            self._metric('ev_departure_success_rate', ev_success_rate, 'District', 'district'),
            self._metric('ev_departure_soc_deficit_mean', ev_deficit_mean, 'District', 'district'),
            self._metric('ev_charge_total_kwh', ev_charge_total, 'District', 'district'),
            self._metric('ev_v2g_export_total_kwh', ev_v2g_total, 'District', 'district'),
        ])

        district_bess_cycles = None if bess_capacity_total <= 0.0 else bess_throughput_total / (2.0 * bess_capacity_total)
        district_bess_fade = None if bess_capacity_total <= 0.0 else bess_capacity_loss_total / bess_capacity_total
        extended_district_rows.extend([
            self._metric('bess_charge_total_kwh', bess_charge_total, 'District', 'district'),
            self._metric('bess_discharge_total_kwh', bess_discharge_total, 'District', 'district'),
            self._metric('bess_throughput_total_kwh', bess_throughput_total, 'District', 'district'),
            self._metric('bess_equivalent_full_cycles', district_bess_cycles, 'District', 'district'),
            self._metric('bess_capacity_fade_ratio', district_bess_fade, 'District', 'district'),
        ])

        district_pv_ratio = None if pv_generation_total <= 0.0 else (pv_generation_total - pv_export_total) / pv_generation_total
        extended_district_rows.extend([
            self._metric('pv_generation_total_kwh', pv_generation_total, 'District', 'district'),
            self._metric('pv_export_total_kwh', pv_export_total, 'District', 'district'),
            self._metric('pv_generation_daily_average_kwh', self._daily_average(pv_generation_total, simulated_days), 'District', 'district'),
            self._metric('pv_export_daily_average_kwh', self._daily_average(pv_export_total, simulated_days), 'District', 'district'),
            self._metric('pv_self_consumption_ratio', district_pv_ratio, 'District', 'district'),
        ])

        district_phase_imbalance = None if phase_imbalance_count <= 0.0 else phase_imbalance_sum / phase_imbalance_count
        extended_district_rows.extend([
            self._metric('electrical_service_violation_total_kwh', phase_violation_total, 'District', 'district'),
            self._metric('electrical_service_violation_time_step_count', phase_violation_count, 'District', 'district'),
            self._metric('phase_imbalance_ratio_average', district_phase_imbalance, 'District', 'district'),
        ])

        phase_union = ['L1', 'L2', 'L3']
        for phase_name in phase_union:
            phase_series = None
            for building in env.buildings:
                history_map = getattr(building, '_charging_phase_power_history_kw', {}) or {}
                if phase_name not in history_map:
                    continue
                t_final = int(max(building.time_step, 0))
                values = np.array(history_map[phase_name][0:t_final + 1], dtype='float64')
                if phase_series is None:
                    phase_series = np.zeros_like(values)
                size = min(len(phase_series), len(values))
                phase_series[:size] += values[:size]

            if phase_series is None:
                continue

            extended_district_rows.append(
                self._metric(f'phase_import_peak_kw_{phase_name}', float(np.clip(phase_series, 0.0, None).max(initial=0.0)), 'District', 'district')
            )
            extended_district_rows.append(
                self._metric(f'phase_export_peak_kw_{phase_name}', float(np.clip(-phase_series, 0.0, None).max(initial=0.0)), 'District', 'district')
            )

        market_by_building, market_district = self._collect_market_totals(building_names)

        for building_name in building_names:
            totals = market_by_building.get(building_name, {})
            local_import = self._to_scalar(totals.get('community_local_import_total_kwh'), 0.0)
            local_export = self._to_scalar(totals.get('community_local_export_total_kwh'), 0.0)
            grid_import = self._to_scalar(totals.get('community_grid_import_after_local_total_kwh'), 0.0)
            grid_export = self._to_scalar(totals.get('community_grid_export_after_local_total_kwh'), 0.0)
            import_share = None if (local_import + grid_import) <= 0.0 else local_import / (local_import + grid_import)
            export_share = None if (local_export + grid_export) <= 0.0 else local_export / (local_export + grid_export)

            extended_building_rows.extend([
                self._metric('community_local_import_total_kwh', local_import, building_name, 'building'),
                self._metric('community_local_export_total_kwh', local_export, building_name, 'building'),
                self._metric('community_grid_import_after_local_total_kwh', grid_import, building_name, 'building'),
                self._metric('community_grid_export_after_local_total_kwh', grid_export, building_name, 'building'),
                self._metric('community_local_import_daily_average_kwh', self._daily_average(local_import, simulated_days), building_name, 'building'),
                self._metric('community_local_export_daily_average_kwh', self._daily_average(local_export, simulated_days), building_name, 'building'),
                self._metric('community_grid_import_after_local_daily_average_kwh', self._daily_average(grid_import, simulated_days), building_name, 'building'),
                self._metric('community_grid_export_after_local_daily_average_kwh', self._daily_average(grid_export, simulated_days), building_name, 'building'),
                self._metric('community_settled_cost_total_eur', totals.get('community_settled_cost_total_eur', 0.0), building_name, 'building'),
                self._metric('community_counterfactual_cost_total_eur', totals.get('community_counterfactual_cost_total_eur', 0.0), building_name, 'building'),
                self._metric('community_market_savings_total_eur', totals.get('community_market_savings_total_eur', 0.0), building_name, 'building'),
                self._metric('community_settled_cost_daily_average_eur', self._daily_average(totals.get('community_settled_cost_total_eur', 0.0), simulated_days), building_name, 'building'),
                self._metric('community_counterfactual_cost_daily_average_eur', self._daily_average(totals.get('community_counterfactual_cost_total_eur', 0.0), simulated_days), building_name, 'building'),
                self._metric('community_market_savings_daily_average_eur', self._daily_average(totals.get('community_market_savings_total_eur', 0.0), simulated_days), building_name, 'building'),
                self._metric('community_local_share_of_demand', import_share, building_name, 'building'),
                self._metric('community_local_share_of_export', export_share, building_name, 'building'),
            ])

        district_local_import = market_district['community_local_import_total_kwh']
        district_local_export = market_district['community_local_export_total_kwh']
        district_grid_import = market_district['community_grid_import_after_local_total_kwh']
        district_grid_export = market_district['community_grid_export_after_local_total_kwh']

        extended_district_rows.extend([
            self._metric('community_local_import_total_kwh', district_local_import, 'District', 'district'),
            self._metric('community_local_export_total_kwh', district_local_export, 'District', 'district'),
            self._metric('community_grid_import_after_local_total_kwh', district_grid_import, 'District', 'district'),
            self._metric('community_grid_export_after_local_total_kwh', district_grid_export, 'District', 'district'),
            self._metric('community_local_import_daily_average_kwh', self._daily_average(district_local_import, simulated_days), 'District', 'district'),
            self._metric('community_local_export_daily_average_kwh', self._daily_average(district_local_export, simulated_days), 'District', 'district'),
            self._metric('community_grid_import_after_local_daily_average_kwh', self._daily_average(district_grid_import, simulated_days), 'District', 'district'),
            self._metric('community_grid_export_after_local_daily_average_kwh', self._daily_average(district_grid_export, simulated_days), 'District', 'district'),
            self._metric('community_settled_cost_total_eur', market_district['community_settled_cost_total_eur'], 'District', 'district'),
            self._metric('community_counterfactual_cost_total_eur', market_district['community_counterfactual_cost_total_eur'], 'District', 'district'),
            self._metric('community_market_savings_total_eur', market_district['community_market_savings_total_eur'], 'District', 'district'),
            self._metric('community_settled_cost_daily_average_eur', self._daily_average(market_district['community_settled_cost_total_eur'], simulated_days), 'District', 'district'),
            self._metric('community_counterfactual_cost_daily_average_eur', self._daily_average(market_district['community_counterfactual_cost_total_eur'], simulated_days), 'District', 'district'),
            self._metric('community_market_savings_daily_average_eur', self._daily_average(market_district['community_market_savings_total_eur'], simulated_days), 'District', 'district'),
            self._metric(
                'community_local_share_of_demand',
                None if (district_local_import + district_grid_import) <= 0.0 else district_local_import / (district_local_import + district_grid_import),
                'District',
                'district',
            ),
            self._metric(
                'community_local_share_of_export',
                None if (district_local_export + district_grid_export) <= 0.0 else district_local_export / (district_local_export + district_grid_export),
                'District',
                'district',
            ),
        ])

        equity_distribution = self._equity_distribution_metrics(np.array(list(equity_valid_benefits.values()), dtype='float64'))
        non_negative_benefits = {name: max(value, 0.0) for name, value in equity_valid_benefits.items()}
        has_complete_manual_groups = all(
            equity_group_by_building.get(name) in {'asset_rich', 'asset_poor'}
            for name in building_names
        )
        equity_bpr = self._equity_bpr(non_negative_benefits, equity_group_by_building) if has_complete_manual_groups else None

        extended_district_rows.extend([
            self._metric('equity_gini_benefit', equity_distribution['equity_gini_benefit'], 'District', 'district'),
            self._metric('equity_cr20_benefit', equity_distribution['equity_cr20_benefit'], 'District', 'district'),
            self._metric('equity_losers_percent', equity_distribution['equity_losers_percent'], 'District', 'district'),
            self._metric('equity_bpr_asset_poor_over_rich', equity_bpr, 'District', 'district'),
        ])

        extended_building = pd.DataFrame(extended_building_rows)
        extended_district = pd.DataFrame(extended_district_rows)

        cost_functions = pd.concat([legacy_cost_functions, extended_district, extended_building], ignore_index=True, sort=False)

        return cost_functions



import os
import re
import setuptools

ROOT = os.path.dirname(__file__)
VERSION_RE = re.compile(r"__version__\s*=\s*['\"]([\w\.-]+)['\"]")
DIST_NAME = os.getenv('CITYLEARN_PYPI_NAME', 'softcpsrecsimulator')
PYPI_README = os.path.join(ROOT, 'README_PYPI.md')

readme_path = PYPI_README if os.path.exists(PYPI_README) else os.path.join(ROOT, 'README.md')

with open(readme_path, 'r', encoding='utf-8') as fh:
    long_description = fh.read()

with open('requirements.txt', 'r', encoding='utf-8') as fh:
   requirements = fh.readlines()
   requirements = [requirement.strip().replace('\n','').replace('\r','') for requirement in requirements]
   requirements = [requirement for requirement in requirements if len(requirement) != 0 and requirement[0] != '#']

def get_version():
    init = open(os.path.join(ROOT, 'citylearn', '__init__.py')).read()
    return VERSION_RE.search(init).group(1)

setuptools.setup(
    name=DIST_NAME,
    version=get_version(),
    author='Soft-CPS Research Group, Jose Ramon Vazquez-Canteli, Kingsley Nweye, Zoltan Nagy',
    author_email='jose@isep.ipp.pt',
    description=(
        'SoftCPS REC Simulator: an energy-community RL simulator fork focused on EV/BESS/PV, '
        'electrical-service constraints, and community-market experimentation.'),
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Soft-CPS-Research-Group/Simulator',
    license='MIT',
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=requirements,
    extras_require={
        'pysam': ['nrel-pysam'],
    },
    entry_points={'console_scripts': ['citylearn = citylearn.__main__:main']},
    project_urls={
        'Source': 'https://github.com/Soft-CPS-Research-Group/Simulator',
        'Original Project': 'https://github.com/intelligent-environments-lab/CityLearn',
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.9',
)



from typing import Any, List, Union
import numpy as np

class Encoder:
    r"""Base class to transform observations.
    """
    
    def __init__(self):
        pass

    def __mul__(self, x: Any):
        raise NotImplementedError

    def __rmul__(self, x: Any):
        raise NotImplementedError

class NoNormalization(Encoder):
    r"""Use to return observation value as-is i.e. without any transformation.

    Examples
    --------
    >>> x_max = 24
    >>> encoder = NoNormalization()
    >>> observation = 2
    >>> encoder*observation
    2
    """
    
    def __init__(self):
        super().__init__()

    def __mul__(self, x: Union[float, int]):
        return x
        
    def __rmul__(self, x: Union[float, int]):
        return x
        
class PeriodicNormalization(Encoder):
    r"""Use to transform observations that are cyclical/periodic e.g. hour-of-day, day-of-week, e.t.c.

    Parameters
    ----------
    x_max : Union[float, int]
        Maximum observation value.

    Notes
    -----
    The transformation returns two values :math:`x_{sin}` and :math:`x_{sin}` defined as:
    
    .. math:: 
        x_{sin} = sin(\frac{2 \cdot \pi \cdot x}{x_{max}})
        
        x_{cos} = cos(\frac{2 \cdot \pi \cdot x}{x_{max}})

    Examples
    --------
    >>> x_max = 24
    >>> encoder = PeriodicNormalization(x_max)
    >>> observation = 2
    >>> encoder*observation
    array([0.75, 0.9330127])
    """
    
    def __init__(self, x_max: Union[float, int]):
        super().__init__()
        self.x_max = x_max

    def __mul__(self, x: Union[float, int]):
        x = 2 * np.pi * x / self.x_max
        x_sin = np.sin(x)
        x_cos = np.cos(x)
        return np.array([x_sin, x_cos])

    def __rmul__(self, x: Union[float, int]):
        x = 2 * np.pi * x / self.x_max
        x_sin = np.sin(x)
        x_cos = np.cos(x)
        return np.array([x_sin, x_cos])

class OnehotEncoding(Encoder):
    r"""Use to transform unordered categorical observations e.g. boolean daylight savings e.t.c.

    Parameters
    ----------
    classes : Union[List[float], List[int], List[str]]
        Observation categories.

    Examples
    --------
    >>> classes = [1, 2, 3, 4]
    >>> encoder = OnehotEncoding(classes)
    # identity_matrix = [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]
    >>> observation = 2
    >>> encoder*observation
    [0, 1, 0, 0]
    """

    def __init__(self, classes: Union[List[float], List[int], List[str]]):
        super().__init__()
        self.classes = classes

    def __mul__(self, x: Union[float, int, str]):
        identity_mat = np.eye(len(self.classes))
        return identity_mat[np.array(self.classes) == x][0]

    def __rmul__(self, x: Union[float, int, str]):
        identity_mat = np.eye(len(self.classes))
        return identity_mat[np.array(self.classes) == x][0]
    
class Normalize(Encoder):
    r"""Use to transform observations to a value between `x_min` and `x_max` using min-max normalization.

    Parameters
    ----------
    x_min : Union[float, int]
        Minimum observation value.
    x_max : Union[float, int]
        Maximum observation value.

    Notes
    -----
    The transformation returns two values :math:`x_{sin}` and :math:`x_{sin}` defined as:
    
    .. math:: 
        x = \frac{x - x_{min}}{x_{max} - x_{min}}

    Examples
    --------
    >>> x_min = 0
    >>> x_max = 24
    >>> encoder = Normalize(x_min, x_max)
    >>> observation = 2
    >>> encoder*observation
    0.08333333333333333
    """
    
    def __init__(self, x_min: Union[float, int], x_max: Union[float, int]):
        super().__init__()
        self.x_min = x_min
        self.x_max = x_max

    def __mul__(self, x: Union[float, int]):
        if self.x_min == self.x_max:
            return 0
        else:
            return (x - self.x_min)/(self.x_max - self.x_min)

    def __rmul__(self, x: Union[float, int]):
        if self.x_min == self.x_max:
            return 0
        else:
            return (x - self.x_min)/(self.x_max - self.x_min)
        
class RemoveFeature(Encoder):
    r"""Use to exlude an observation by returning `None` type.

    Examples
    --------
    >>> encoder = RemoveFeature()
    >>> observation = 2
    >>> encoder*observation
    None
    """
    
    def __init__(self):
        super().__init__()
        pass

    def __mul__(self, x: Any):
        return None

    def __rmul__(self, x: Any):
        return None